/*
 * main.cpp
 *
 *  Created on: Aug 10, 2012
 *      Author: scottyoung
 */

#include <stdio.h>

#import "scene_parser.h"
#import "hit.h"
#import "image.h"
#import "math.h"

//Some rendering constants
#define D_LO 12
#define D_HI 17

//Different shading options
Vec3f flatShading(Hit *h, SceneParser *theScene){return(h->getMaterial()->color);}
Vec3f depthShading(Hit *h, SceneParser *theScene){
	float depth = 1.0 - (h->getT()-D_LO)/(D_HI-D_LO);
	Vec3f grey(depth, depth, depth);
	return(grey);
}
Vec3f normalShading(Hit *h, SceneParser *theScene){Vec3f N = h->getNormal(); N.Abs(); return(N);}
Vec3f diffuseShading(Hit *h, SceneParser *theScene){
	int nL = theScene->getNumLights();
	Light **lights = theScene->getLights();
	Vec3f ambient = theScene->getAmbientLight();
	Vec3f color = h->getMaterial()->color;
	Vec3f value(color);
	value.Scale(ambient.r(), ambient.g(), ambient.b());
	for(int i=0; i<nL; i++)
	{
		Vec3f I, dir;
		lights[i]->getIllumination(h->getIntersectionPoint(), dir, I);
		float d = dir.Dot3(h->getNormal());
		if(d>0)
		{
			Vec3f additiveLight(I.r()*color.r()*d, I.g()*color.g()*d, I.b()*color.b()*d);
			value += additiveLight;
		}
	}
	return(value);
}

Image render(int resolution, SceneParser *theScene, float t_min, Vec3f (*renderOption)(Hit*, SceneParser*))
{
	Image img(resolution, resolution);
	Camera *cam = theScene->getCamera();
	Group *group = theScene->getGroup();
	Vec3f background = theScene->getBackgroundColor();

	img.SetAllPixels(background);
	for(int x=0; x<resolution; x++)
	{
		for(int y=0; y<resolution; y++)
		{
			Vec2f pos((float)x/(resolution-1.0), (float)y/(resolution-1.0));
			Ray r = cam->generateRay(pos);
			Hit h;
			if(group->intersect(r,h,t_min))
			{
				Vec3f color = renderOption(&h, theScene);
				img.SetPixel(x, y, color);
			}
		}
	}
	return(img);
}

/**
 * This loads a file which will consecutively render each scene file
 * and output to the given image file, while providing rendering updates
 * (resolution is identical and specified by RESOLUTION)
 */
void renderAll(const char* filename)
{
	const int RESOLUTION = 512;
	assert(filename != NULL);
	const char *ext = &filename[strlen(filename)-4];
	assert(!strcmp(ext,".txt"));
	FILE *file = fopen(filename,"r");
	assert (file != NULL);

	char token[MAX_PARSER_TOKEN_LENGTH];
	while (fscanf(file, "%s", token) != EOF)
	{
		if(!strcmp(token, "RENDER"))
		{
			char input[MAX_PARSER_TOKEN_LENGTH];
			char output[MAX_PARSER_TOKEN_LENGTH];
			fscanf(file, "%s", input);
			fscanf(file, "%s", output);
			cout << "RENDERING " << input << " --> " << output << endl;
			SceneParser *theScene = new SceneParser(input);
			Image image = render(RESOLUTION, theScene, 0, diffuseShading);
			image.SaveTGA(output);
			delete theScene;
		}
	}

}

//Used for rendering only one scene
void renderOne()
{
	SceneParser theScene("scene11_axes_cube.txt");
	Group *theGroup = theScene.getGroup();
	Camera *theCamera = theScene.getCamera();
	Light **theLights = theScene.getLights();
	int numLights = theScene.getNumLights();
	Vec3f ambient = theScene.getAmbientLight();
	Vec3f background = theScene.getBackgroundColor();

	theCamera->print();
	cout << "Ambient Light " << "R: " << ambient.r() << ", G: " << ambient.g()
			<< ", B: " << ambient.b() << endl;
	cout << "Background " << "R: " << background.r() << ", G: " << background.g()
			<< ", B: " << background.b() << endl;
	for(int i=0; i<numLights; i++)
		theLights[i]->print();
	theGroup->print();

	Image scene = render(512, &theScene, 0, diffuseShading);
	scene.SaveTGA("scene11.tga");
}

int main()
{
	renderAll("scene_list.txt");
	return(0);
}



