/*
 * main.cpp
 *
 *  Created on: Aug 10, 2012
 *      Author: scottyoung
 */

#include <stdio.h>

#import "scene_parser.h"
#import "hit.h"
#import "image.h"
#import "math.h"
#import "RayTracer.h"

//Some rendering constants
#define D_LO 11
#define D_HI 15
#define MAX_BOUNCES 5
#define MIN_WEIGHT 0.01
#define GRID_DIMENSION 100
#define CAST_SHADOWS 1

//Different shading options
Vec3f flatShading(const Ray &r, Hit *h, SceneParser *theScene, float tMin){return(h->getMaterial()->flatColor());}
Vec3f depthShading(const Ray &r, Hit *h, SceneParser *theScene, float tMin){
	float depth = 1.0 - (h->getT()-D_LO)/(D_HI-D_LO);
	Vec3f grey(depth, depth, depth);
	return(grey);
}
Vec3f normalShading(const Ray &r, Hit *h, SceneParser *theScene, float tMin){Vec3f N = h->getNormal(); N.Abs(); return(N);}
//Performs only diffuse shading (no specular, transparent, reflective)
Vec3f diffuseShading(const Ray &r, Hit *h, SceneParser *theScene, float tMin){
	int nL = theScene->getNumLights();
	Light **lights = theScene->getLights();
	Vec3f ambient = theScene->getAmbientLight();
	Vec3f color = h->getMaterial()->flatColor();
	Vec3f value(color);
	value.Scale(ambient.r(), ambient.g(), ambient.b());
	for(int i=0; i<nL; i++)
	{
		Vec3f I, dir;
		float dummyFloat;
		lights[i]->getIllumination(h->getIntersectionPoint(), dir, I, dummyFloat);
		float d = dir.Dot3(h->getNormal());
		if(d>0)
		{
			Vec3f additiveLight(I.r()*color.r()*d, I.g()*color.g()*d, I.b()*color.b()*d);
			value += additiveLight;
		}
	}
	return(value);
}
//Shades based on specular and diffuse components
Vec3f phongShading(const Ray &r, Hit *h, SceneParser *theScene, float tMin)
{
	int nL = theScene->getNumLights();
	Light **lights = theScene->getLights();
	Vec3f ambient = theScene->getAmbientLight();
	Material *m = h->getMaterial();
	Vec3f value(m->flatColor());
	value.Scale(ambient.r(), ambient.g(), ambient.b());
	for(int i=0; i<nL; i++)
	{
		Vec3f I, dir;
		float dummyFloat;
		lights[i]->getIllumination(h->getIntersectionPoint(), dir, I, dummyFloat);
		Vec3f phong = m->shade(r,h,dir,I);
		value += phong;
	}
	return(value);
}
//Full ray tracer which casts shadows, reflection, refraction and specular/diffuse shading
Vec3f rayTrace(const Ray &r, Hit *h, SceneParser *theScene, float tMin)
{
	RayTracer *rayTrace = new RayTracer(theScene, MAX_BOUNCES, MIN_WEIGHT, CAST_SHADOWS);
	return(rayTrace->traceRay(r, tMin, 0, 1.0, 1.0, *h));
}

Image render(int resolution, SceneParser *theScene, float t_min, Vec3f (*renderOption)(const Ray &, Hit*, SceneParser*, float), Grid *grid = 0)
{
	Image img(resolution, resolution);
	Camera *cam = theScene->getCamera();
	Group *group = theScene->getGroup();
	Vec3f background = theScene->getBackgroundColor();

	img.SetAllPixels(background);
	for(int x=0; x<resolution; x++)
	{
		for(int y=0; y<resolution; y++)
		{
			//if(x==250 && y==450)//DEBUG
			//{
				Vec2f pos((float)x/(resolution-1.0), (float)y/(resolution-1.0));
				Ray r = cam->generateRay(pos);
				Hit h;

				RayTracingStats::IncrementNumNonShadowRays();

				bool wasHit = false;
				if(grid != 0)
					wasHit = grid->intersect(r,h,t_min);
				else
					wasHit = group->intersect(r,h,t_min);

				if(wasHit)
				{
					Vec3f color = renderOption(r, &h, theScene, t_min);
					img.SetPixel(x, y, color);
				}
			//}
		}
	}
	return(img);
}

Image renderFast(int resolution, SceneParser *theScene, float t_min, Vec3f (*renderOption)(const Ray &, Hit*, SceneParser*, float), Grid *grid = 0){

		Group *theGroup = theScene->getGroup();
		Grid *accelGrid = Grid::make(theGroup, GRID_DIMENSION, GRID_DIMENSION, GRID_DIMENSION);
		Group *newGroup = new Group(1);
		newGroup->addObject(accelGrid);
		theScene->setGroup(newGroup); //Make the group the grid so that raytracing works properly

		return(render(resolution, theScene, t_min, renderOption));

}

/**
 * This loads a file which will consecutively render each scene file
 * and output to the given image file, while providing rendering updates
 * (resolution is identical and specified by RESOLUTION)
 */
void renderAll(const char* filename)
{
	const int RESOLUTION = 512;
	assert(filename != NULL);
	const char *ext = &filename[strlen(filename)-4];
	assert(!strcmp(ext,".txt"));
	FILE *file = fopen(filename,"r");
	assert (file != NULL);

	char token[MAX_PARSER_TOKEN_LENGTH];
	while (fscanf(file, "%s", token) != EOF)
	{
		if(!strcmp(token, "RENDER"))
		{
			char input[MAX_PARSER_TOKEN_LENGTH];
			char output[MAX_PARSER_TOKEN_LENGTH];
			fscanf(file, "%s", input);
			fscanf(file, "%s", output);
			cout << "RENDERING " << input << " --> " << output << endl;
			SceneParser *theScene = new SceneParser(input);
			Group *theGroup = theScene->getGroup();
			RayTracingStats::Initialize(RESOLUTION, RESOLUTION, GRID_DIMENSION, GRID_DIMENSION, GRID_DIMENSION, theGroup->getBoundingBox().getMin(), theGroup->getBoundingBox().getMax());
			Image image = renderFast(RESOLUTION, theScene, 0, rayTrace);
			image.SaveTGA(output);
			RayTracingStats::PrintStatistics();
			delete theScene;
		}
	}

}

//Used for rendering only one scene
void renderOne(const char * theFile, const char * theImage)
{
	const int RESOLUTION = 512;
	SceneParser theScene(theFile);
	Group *theGroup = theScene.getGroup();
	Camera *theCamera = theScene.getCamera();
	Light **theLights = theScene.getLights();
	Material **theMaterials = theScene.getMaterials();
	int numLights = theScene.getNumLights();
	int numMaterials = theScene.getNumMaterials();
	Vec3f ambient = theScene.getAmbientLight();
	Vec3f background = theScene.getBackgroundColor();

	RayTracingStats::Initialize(RESOLUTION, RESOLUTION, GRID_DIMENSION, GRID_DIMENSION, GRID_DIMENSION, theGroup->getBoundingBox().getMin(), theGroup->getBoundingBox().getMax());

	theCamera->print();
	cout << "Ambient Light " << "R: " << ambient.r() << ", G: " << ambient.g()
			<< ", B: " << ambient.b() << endl;
	cout << "Background " << "R: " << background.r() << ", G: " << background.g()
			<< ", B: " << background.b() << endl;
	for(int i=0; i<numLights; i++)
		theLights[i]->print();
	theGroup->print();
	cout << "Materials: " << endl;
	for(int i=0; i<numMaterials; i++)
		theMaterials[i]->print();


	Image scene = renderFast(RESOLUTION, &theScene, 0, rayTrace);
	scene.SaveTGA(theImage);
}

int main()
{
	cout << "--PROGRAM STARTED--" << endl;
	renderAll("scenelist.txt");
	cout << "--PROGRAM COMPLETE--" << endl;
}



